﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using Microsoft.AnalysisServices.AdomdClient;
using System.Runtime.Serialization.Formatters.Binary;

namespace Tests
{
    public class BaseMDXTest
    {
        private string connectionstring;
        private string baseOutputDir;
        private string outputDir;
        private Boolean recreateTemplate = false;
        private static string testReportHtml;
        private static int numPassed = 0;
        private static int numFailed = 0;

        /** Construct BaseMDX Test
         * connectionstring - The URL of the Analysis Server on which the MDX query is executed
         * baseOutputDir - The output directory under which the MDX query results are stored
         * recreateTemplate - Force the recreation of the MDX query result template
         **/
        public BaseMDXTest(string connectionstring, string baseOutputDir, Boolean recreateTemplate = false)
        {
            this.connectionstring = connectionstring;
            this.baseOutputDir = baseOutputDir;
            this.recreateTemplate = recreateTemplate;
            if (testReportHtml == null)
                testReportHtml = "<html><h1>Test Report</h1><table border=\"1\"><tr><th>Test Case</th><th>Pass/Fail</th></tr>";
        }

        /** Open a connection to the Analysis Services database **/
        static AdomdConnection GetAdomdConnection(string connectionstring)
        {
            var connection = new AdomdConnection(connectionstring);
            connection.Open();
            return connection;
        }

        /** Runs the given MDX query, creates a template if it doesn't exist, and stores the results 
         * mdxTestTitle - The name of the unit-test
         * mdxQuery - The MDX query to test
         **/
        public Boolean TestMDXQuery(string mdxTestTitle, string mdxQuery)
        {
            Dictionary<string, CubeRow> templateCubeDict = new Dictionary<string, CubeRow>();
            string templateHtml = null;
            string resultHtml = "<html><h2>Test Title</h2><p>" + mdxTestTitle + "</p><h2>MDX Query</h2><p>" + mdxQuery + "</p>\n";
            // If the template for the given test already exists and template recreation has not been requested
            // then run the comparison.
            if (!recreateTemplate && TemplateExists(mdxTestTitle))
            {
                resultHtml += "<h2>Original Template</h2>\n";
                // Lots the origional template HTML
                templateHtml = loadTemplate(mdxTestTitle, templateCubeDict);
                resultHtml += templateHtml + "\n";
                resultHtml += "<h2>Test Results</h2>\n";
                // Run the MDX query and compare against the template
                CubeComparisonResult result = RunComparison(mdxQuery, templateCubeDict);
                resultHtml += result.getResultHtml() + "\n";
                resultHtml += result.hasPassed() ? "<h2 stlye=\"color:green\";>Test Passes!</h2>" : "<h2 style=\"color:red;\">Test Fails</h2>";
                resultHtml += "</html>";
                // Persist the MDX query result
                string testReportFile = PersistResult(mdxTestTitle, resultHtml);
                if (result.hasPassed())
                    numPassed++;
                else
                    numFailed++;
                testReportHtml += "<tr><td><a href=\"file:\\" + testReportFile + "\">" + mdxTestTitle + "</a></td><td>" + (result.hasPassed() ? "Pass" : "Fail") + "</td></tr>";
                // Persist the HTML summary report
                PersistTestReport(numPassed, numFailed);
                return result.hasPassed();
            }
            else // Else build the MDX query result template if the template does not previously exist or a request to force the re-creation of the template.
            {
                resultHtml += "<h2>Original Template</h2>\n";
                // Build the template using the MDXQuery and store it with the Dictionary<string, CubeRow>
                templateHtml = BuildTemplate(mdxQuery, templateCubeDict);
                resultHtml += templateHtml + "\n";
                resultHtml += "<h2 style=\"color:red\";>Test Fails.  No Template exists.  You must validate the output by hand</h2>";
                resultHtml += "</html>";
                // Persist the template.  This includes the Dictionary<string, CubeRow> and the HTML version of the MDX query result
                string resultHtmlFile = PersistTemplate(mdxTestTitle, templateHtml, templateCubeDict);
                testReportHtml += "<tr><td><a href=\"file:\\" + resultHtmlFile + "\">" + mdxTestTitle + "</a></td><td>Fail - Template Did not Exist</td></tr>";
                // Creation of the template is always considered a failure since the unit-test developer must hand validate
                // the results before the template is considered usable.
                numFailed++;
                // Persist the HTML summary report
                PersistTestReport(numPassed, numFailed);
                return false;
            }
        }

        /** Save a summary of the test results **/
        private void PersistTestReport(int numPassed, int numFailed)
        {
            string testReportFile = baseOutputDir + "\\TestReport.html";
            string tempReport = testReportHtml;
            tempReport += "</table>";
            tempReport += "<h3>Num Passed: " + numPassed + "</h3>";
            tempReport += "<h3>Num Failed: " + numFailed + "</h3>";
            tempReport += "</html>";
            using (StreamWriter sw = File.CreateText(testReportFile))
            {
                sw.WriteLine(tempReport);
            }
        }

        /** Load both the HTML and Object version of the MDX query result template **/
        private string loadTemplate(string mdxTestTitle, Dictionary<string, CubeRow> templateCubeDict)
        {
            string templateFile = GetTemplateDir(mdxTestTitle) + "\\template.html";
            string templateHtml = null;
            using (StreamReader sr = File.OpenText(templateFile))
            {
                templateHtml = sr.ReadToEnd();
            }

            string cubeDictFile = GetTemplateDir(mdxTestTitle) + "\\CubeDict";
            using (Stream stream = File.Open(cubeDictFile, FileMode.Open))
            {
                BinaryFormatter binaryFormatter = new BinaryFormatter();
                Dictionary<string, CubeRow> templateDict = (Dictionary<string, CubeRow>)binaryFormatter.Deserialize(stream);
                foreach (var cubeRow in templateDict)
                {
                    templateCubeDict.Add(cubeRow.Key, cubeRow.Value);
                }
            }
            return templateHtml;
        }

        /** Save the HTML file that contains the MDX query comparison results **/
        private string PersistResult(string mdxTestTitle, string resultHtml)
        {
            string outputDir = baseOutputDir + "\\result\\" + mdxTestTitle;
            if (!System.IO.Directory.Exists(@outputDir))
                System.IO.Directory.CreateDirectory(@outputDir);

            string resultHtmlFile = outputDir + "\\result.html";
            using (StreamWriter sw = File.CreateText(resultHtmlFile))
            {
                sw.WriteLine(resultHtml);
            }
            return resultHtmlFile;
        }

        /** Save the HTML file and Dictionary<string, CubeRow> Object that contains the initial version of the
         * MDX query results.
         * mdxTestTitle - The name of the unit-test
         * resultHTML - The HTML representation of the MDX query results.
         * templateCubeDict - THe object representation of the MDX query results.
         **/
        private string PersistTemplate(string mdxTestTitle, string resultHtml, Dictionary<string, CubeRow> templateCubeDict)
        {
            string templateDir = GetTemplateDir(mdxTestTitle);
            if (!System.IO.Directory.Exists(@templateDir))
                System.IO.Directory.CreateDirectory(@templateDir);

            string resultHtmlFile = templateDir + "\\template.html";
            using (StreamWriter sw = File.CreateText(resultHtmlFile))
            {
                sw.WriteLine(resultHtml);
            }

            string cubeDictFile = templateDir + "\\CubeDict";
            using (Stream stream = File.Open(cubeDictFile, FileMode.Create))
            {
                BinaryFormatter formatter = new BinaryFormatter();
                formatter.Serialize(stream, templateCubeDict);
            }
            return resultHtmlFile;
        }

        /** 
         * Compares the results of a MDX query with the previously peristed results
         **/
        private CubeComparisonResult RunComparison(string mdxQuery, Dictionary<string, CubeRow> templateCubeDict)
        {
            string tableResult = "<table border=\"1\">";
            Boolean testPass = true;
            // Open a connection to the Analysis server
            using (var connection = GetAdomdConnection(connectionstring))
            // Execute the MDX query
            using (var command = new AdomdCommand(mdxQuery, connection))
            {
                // Get the MDX query results as a CellSet object
                CellSet cst = command.ExecuteCellSet();

                // Make sure that the result contains at least two axes.  Columns and rows.
                int axesCount = cst.Axes.Count;
                if (axesCount != 2)
                    throw new Exception("The MDX Query did not return two axes.  Please check your query.");

                // Make sure that data is contained within the two axes.
                for (int count = 0; count < axesCount; count++)
                {
                    if (cst.Axes[count].Positions.Count == 0)
                        throw new Exception("The MDX query return no data.  Please check your query.");
                }

                // Get the number of column dimensions
                int colDimCount = cst.Axes[0].Positions[0].Members.Count;
                // Get the number of row dimensions
                int rowDimCount = 0;
                if (cst.Axes[1].Positions[0].Members.Count > 0)
                    rowDimCount = cst.Axes[1].Positions[0].Members.Count;

                // The row count is the total number of Position(s) on the row axis plus the number of column dimensions
                // The reason the column dimension count is included is that the column headers themselves make up rows
                int rowCount = cst.Axes[1].Positions.Count + colDimCount;
                // The column count is the total number of Position(s) on the column axis plus the number of row dimensions.
                // The reason the row dimension count is included is that the row headers themselves make up columns
                int colCount = cst.Axes[0].Positions.Count + rowDimCount;

                // Create an array of string(s) that hold the column headers
                string[] columnHeader = new string[cst.Axes[0].Positions.Count];

                // Create a point to hold the current CubeRow dictionary
                Dictionary<string, CubeRow> currentCubeDict = templateCubeDict;
                // Create a pointer to hold the current CubeRow object
                CubeRow currentCubeRow = null;
                // Loop through all the rows within the MDX result
                for (int currentRow = 0; currentRow < rowCount; currentRow++)
                {
                    // Create a new HTML table row
                    tableResult += "<tr>";

                    // If the current row is greater than the number of column dimensions then 
                    // processing of actual row data is occurring.  Set the CubeRow pointer to a new
                    // CubeRow and set the CubeRow Dictionary pointer to the original Dictionary passed in.
                    // This is also the head of the Dictionary structure.
                    if (currentRow >= colDimCount)
                    {
                        currentCubeRow = new CubeRow();
                        currentCubeDict = templateCubeDict;
                    }

                    // Loop through all the columns within the MDX result
                    for (int currentCol = 0; currentCol < colCount; currentCol++)
                    {
                        // Check if the current row is dealing with column dimension captions
                        if (currentRow < colDimCount)
                        {
                            if (currentCol < rowDimCount)
                            {
                                // If the current column is has not reached the row dimension headers create a blank HTML cell
                                // These cells make up the uper left hand corner of the HTML table as place holders before the 
                                // HTML column headers and row headers meet.
                                tableResult += "<th> </th>";
                            }
                            else
                            {
                                // A column header is being processed.  The column header may be many levels deep depending on
                                // how many dimensions are included in the column portion of the MDX query.  The location of the column header is 
                                // computed by taking the current column count minus the number of row dimensions since getting the total number of 
                                // columns involved adding the total number of row dimensions.

                                // Get the current column index which is the current column count - the number of row dimensions since
                                // the row dimensions captions are part of the columns.
                                int columnIndex = currentCol - rowDimCount;

                                // If the column header at the given location does not exist add it.
                                if (columnHeader[columnIndex] == null)
                                    columnHeader[columnIndex] = cst.Axes[0].Positions[columnIndex].Members[currentRow].Caption;
                                else
                                    // Otherwise if the column header does exist append to it.
                                    columnHeader[columnIndex] += ":" + cst.Axes[0].Positions[columnIndex].Members[currentRow].Caption;

                                // Create an HTML table header cell to hold the column name and add it to the HTML template.
                                tableResult += "<th>" + cst.Axes[0].Positions[columnIndex].Members[currentRow].Caption + "</th>";
                            }
                        }
                        else
                        {
                            // Each cell of the HTML table will be colored green (to indicate a match with the template) or red (to indicate a mismatch)
                            // Always the optomist the cell color is initialized as green.
                            string color = "green";
                            if (currentCol < rowDimCount)
                            {
                                // Get the current row index which is the current row count - the number of column dimensions since the 
                                // column dimension captions are part of the rows.
                                int rowIndex = currentRow - colDimCount;                                
                                // Capture the row caption since processing at this point is not within the row values
                                string rowCaption = cst.Axes[1].Positions[rowIndex].Members[currentCol].Caption;

                                if (!currentCubeDict.ContainsKey(rowCaption))
                                {
                                    // If the template does not contain the row caption fail the test and set the color to red.
                                    // Color all cells within this row red since this is extra data that is not in the template.
                                    color = "red";
                                    testPass = false;
                                }
                                else
                                {
                                    // If the template contains the row caption extract the CubeRow dictionary that
                                    // applies to this row caption and point the current CubeRow and CubeRow Dictionary 
                                    // to the extracted CubeRow.
                                    currentCubeRow = currentCubeDict[rowCaption];
                                    currentCubeDict = currentCubeRow.GetCubeDict();
                                }
                                // Add the row caption to the HTML table appropriately colored.
                                tableResult += "<th bgcolor=\"" + color + "\">" + rowCaption + "</th>";
                            }
                            else
                            {
                                // Get the current column index which is the current column count - the number of row dimensions since
                                // the row dimensions captions are part of the columns.
                                int columnIndex = currentCol - rowDimCount;
                                // Get the current row index which is the current row count - the number of column dimensions since the 
                                // column dimension captions are part of the rows.
                                int rowIndex = currentRow - colDimCount;
                                // Get the current column name for the array of column headers
                                string columnName = columnHeader[columnIndex];
                                // Get the current column value based on the current column and row
                                string value = cst[columnIndex, rowIndex].FormattedValue;
                                //if (value.Equals("621369"))
                                //    value = "621370";
                                if (currentCubeRow == null)
                                {
                                    // if the current cube row does not exist color all cells red and fail the test.
                                    color = "red";
                                    testPass = false;
                                }
                                else
                                {
                                    // Get the CubeTuple for the the current column name
                                    CubeTuple templateValue = currentCubeRow.GetCubeTuple(columnName);
                                    if (templateValue == null)
                                    {
                                        // If the column name returned by the MDX query does not exist in the template.
                                        // then the test is considered to have failed.
                                        color = "red";
                                        testPass = false;
                                    }
                                    else if (!templateValue.getColumnValue().Equals(value))
                                    {
                                        // If the column name does exist within the tempalte but the value associated with
                                        // the column name does not equal what is in the tempplate then the test is 
                                        // considered to have failed.
                                        color = "red";
                                        testPass = false;
                                    }
                                    // Remove the CubeTuple for the current column name from the template data structure.
                                    currentCubeRow.RemoveCubeTuple(columnName);
                                }
                                tableResult += "<td white-space:nowrap bgcolor=\"" + color + "\">" + value + "</td>";
                            }
                        }
                    }
                    tableResult += "</tr>";
                }
                tableResult += "</table>";

                if (testPass)
                {
                    // If the tests thus far has passed check to see if there is still CubeTuple(s) that exist within the
                    // template.  If there are then extra data exists in the template that was not returned by the current
                    // MDX query.
                    testPass = !CheckForExtraDataInTemplate(templateCubeDict);
                }
                return new CubeComparisonResult(testPass, tableResult);
            }
        }

        /** Recursively checks the MDX query result dictionary for extra data that was not found
         * during a subsequent run.  What this means is that the test fails because it does not have as much 
         * data returned as the origional query.
         **/
        private Boolean CheckForExtraDataInTemplate(Dictionary<string, CubeRow> templateCubeDict)
        {
            foreach (string key in templateCubeDict.Keys)
            {
                CubeRow cubeRow = templateCubeDict[key];
                if (cubeRow.GetCubeDict() != null)
                    return CheckForExtraDataInTemplate(cubeRow.GetCubeDict());
                else
                {
                    if (cubeRow.GetCubeTupleSize() > 0)
                        return true;
                }
            }
            return false;
        }

        /** Builds the HTML and Dictionary<string, CubeRow> MDX query result objects give an MDX query 
         * mdxQuery - The MDX query that contains the test
         * templateCubeDict - The datastructure that holds the MDX query results
         */
        private string BuildTemplate(string mdxQuery, Dictionary<string, CubeRow> templateCubeDict)
        {
            string tableResult = "<table border=\"1\">";
            // Open a connection to the Analysis server
            using (var connection = GetAdomdConnection(connectionstring))
            // Execute the MDX query
            using (var command = new AdomdCommand(mdxQuery, connection))
            {
                // Get the MDX query results as a CellSet object
                CellSet cst = command.ExecuteCellSet();
                // Make sure that the result contains at least two axes.  Columns and rows.
                int axesCount = cst.Axes.Count;
                if (axesCount != 2)
                    throw new Exception("The MDX Query did not return two axes.  Please check your query.");

                // Make sure that data is contained within the two axes.
                for (int count = 0; count < axesCount; count++)
                {
                    if (cst.Axes[count].Positions.Count == 0)
                        throw new Exception("The MDX query return no data.  Please check your query.");
                }

                // Get the number of column dimensions
                int colDimCount = cst.Axes[0].Positions[0].Members.Count;
                // Get the number of row dimensions
                int rowDimCount = 0;
                if (cst.Axes[1].Positions[0].Members.Count > 0)
                    rowDimCount = cst.Axes[1].Positions[0].Members.Count;

                // The row count is the total number of Position(s) on the row axis plus the number of column dimensions
                // The reason the column dimension count is included is that the column headers themselves make up rows
                int rowCount = cst.Axes[1].Positions.Count + colDimCount;
                // The column count is the total number of Position(s) on the column axis plus the number of row dimensions.
                // The reason the row dimension count is included is that the row headers themselves make up columns
                int colCount = cst.Axes[0].Positions.Count + rowDimCount;

                // Create an array of string(s) that hold the column headers
                string[] columnHeader = new string[cst.Axes[0].Positions.Count];

                // Create a pointer to hold the current CubeRow object
                CubeRow currentCubeRow = null;
                // Create a point to hold the current CubeRow dictionary
                Dictionary<string, CubeRow> currentCubeDict = templateCubeDict;

                // Loop through all the rows within the MDX result
                for (int currentRow = 0; currentRow < rowCount; currentRow++)
                {
                    // Create a new HTML table row
                    tableResult += "<tr>";

                    // If the current row is greater than the number of column dimensions then 
                    // processing of actual row data is occurring.  Set the CubeRow pointer to a new
                    // CubeRow and set the CubeRow Dictionary pointer to the original Dictionary passed in.
                    // This is also the head of the Dictionary structure.
                    if (currentRow >= colDimCount)
                    {
                        currentCubeRow = new CubeRow();
                        currentCubeDict = templateCubeDict;
                    }

                    // Loop through all the columns within the MDX result
                    for (int currentCol = 0; currentCol < colCount; currentCol++)
                    {
                        // Check if the current row is dealing with column dimension captions
                        if (currentRow < colDimCount)
                        {
                            if (currentCol < rowDimCount)
                            {
                                // If the current column is has not reached the row dimension headers create a blank HTML cell
                                // These cells make up the uper left hand corner of the HTML table as place holders before the 
                                // HTML column headers and row headers meet.
                                tableResult += "<th> </th>";
                            }
                            else
                            {
                                // A column header is being processed.  The column header may be many levels deep depending on
                                // how many dimensions are included in the column portion of the MDX query.  The location of the column header is 
                                // computed by taking the current column count minus the number of row dimensions since getting the total number of 
                                // columns involved adding the total number of row dimensions.

                                // Get the current column index which is the current column count - the number of row dimensions since
                                // the row dimensions captions are part of the columns.
                                int columnIndex = currentCol - rowDimCount;

                                // If the column header at the given location does not exist add it.
                                if (columnHeader[columnIndex] == null)
                                    columnHeader[columnIndex] = cst.Axes[0].Positions[columnIndex].Members[currentRow].Caption;
                                else
                                    // Otherwise if the column header does exist append to it.
                                    columnHeader[columnIndex] += ":" + cst.Axes[0].Positions[columnIndex].Members[currentRow].Caption;

                                // Create an HTML table header cell to hold the column name and add it to the HTML template.
                                tableResult += "<th>" + cst.Axes[0].Positions[columnIndex].Members[currentRow].Caption + "</th>";
                            }
                        }
                        else // If current row is greater then or equal to the column dimension count then processing of the MDX row data commences
                        {
                            // If the current column count is less than the row count then a row header is being traversed.
                            if (currentCol < rowDimCount)
                            {
                                if (currentCol > 0)
                                {
                                    // If the current column count is greater than 0 then a new row header is being processed.
                                    // Create a new CubeRow Dictionary to store the contents of the given row header.
                                    if (currentCubeRow.GetCubeDict() == null)
                                        currentCubeRow.setCubeDict(new Dictionary<string, CubeRow>());
                                    currentCubeDict = currentCubeRow.GetCubeDict();
                                    currentCubeRow = new CubeRow();
                                }

                                // Get the current row index which is the current row count - the number of column dimensions since the 
                                // column dimension captions are part of the rows.
                                int rowIndex = currentRow - colDimCount;
                                // Store the row header caption
                                string rowCaption = cst.Axes[1].Positions[rowIndex].Members[currentCol].Caption;
                                if (!currentCubeDict.ContainsKey(rowCaption))
                                {
                                    // If the row header caption does not exist in the current CubeRow dictionary add it and the 
                                    // new CubeRow
                                    currentCubeDict.Add(rowCaption, currentCubeRow);
                                }
                                else
                                {
                                    // if the row header does exist in the current CubeRow dictionary then use it to retrieve the current
                                    // CubeRow value.
                                    currentCubeRow = currentCubeDict[rowCaption];
                                }

                                // Update the HTML to include the row header caption
                                tableResult += "<th>" + rowCaption + "</th>";
                            }
                            else
                            {
                                // Get the current column index which is the current column count - the number of row dimensions since
                                // the row dimensions captions are part of the columns.
                                int columnIndex = currentCol - rowDimCount;
                                // Get the current row index which is the current row count - the number of column dimensions since the 
                                // column dimension captions are part of the rows.
                                int rowIndex = currentRow - colDimCount;
                                // Add a CubeTuple at the current column and row location with the current column header and Cube value
                                currentCubeRow.AddTuple(columnHeader[columnIndex], cst[columnIndex, rowIndex].FormattedValue);
                                // Add the Cube cell value within an HTML cell.
                                tableResult += "<td white-space:nowrap>" + cst[columnIndex, rowIndex].FormattedValue + "</td>";
                            }
                        }
                    }
                    tableResult += "</tr>";
                }
                tableResult += "</table>";

                return tableResult;
            }
        }

        /** Check if the MDX query result template with the given MDX test name exists **/
        private bool TemplateExists(string mdxTestTitle)
        {
            string templateDir = GetTemplateDir(mdxTestTitle);
            return System.IO.Directory.Exists(templateDir);
        }

        /** Returns the template directory for the given MDX test name **/
        private string GetTemplateDir(string mdxTestTitle)
        {
            return baseOutputDir + "\\templates\\" + mdxTestTitle;
        }
    }
}
