using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.AnalysisServices;
using System.IO;
using System.Xml;

namespace Tests
{
    public class XMLAtoCube
    {
        private string cubeConfigBaseDir;

        public XMLAtoCube(string cubeConfigBaseDir)
        {
            this.cubeConfigBaseDir = cubeConfigBaseDir;
        }

        /** Connect to an Analysis server using the given connection string
         * connectionstring - Anaylsis server URL
         */
        private Server ConnectToServer(string connectionstring)
        {
            Server server = new Server();
            server.Connect(connectionstring);
            return server;
        }

        /** Drops the a database on the connected Analysis server 
         * server - The connected Analysis server
         * databaseName - The name of the unit-test version of the database
         */
        private void DropDatabase(Server server, string databaseName)
        {
            Database db = null;
            if (server.Connected)
            {
                // Find the Database instance given its name
                db = server.Databases.FindByName(databaseName);
                if (db != null)
                {
                    // If the database already exists that its dropped
                    db.Drop();
                }
            }
        }

        /** Retrieve the datbase instance on the connected Analysis serer given the database's name
         * server - The connected Analysis server
         * databaseName - The name of the unit-test version of the database
         */
        private Database GetDatabase(Server server, string databaseName)
        {
            Database db = null;
            if (server.Connected)
            {
                db = server.Databases.FindByName(databaseName);
            }
            return db;
        }

        /** Creates the unit-test version of the Cube given the production version's XMLA definition.  The unit test
         * version is created on the specified target Analysis server and database.
         * baseDir - The base directory where the XMLA definition is stored.
         * sourceServerName - The name of the server where the production version of the Cube resides.
         * sourceDatabaseName - The name of the database where the prodution version of the Cube resides.
         * targetDatabaseName - The name of the unit-test version of the Cube database.
         * dataSourceProviderDefinition - The connection string URL pointing to the test version of the source dimensions and fact tables.
         * targetServer - The name of the unit-test Analysis server.
         */
        private void AddCubeToDatabase(string baseDir, string sourceServerName, string sourceDatabaseName, string targetDatabaseName, string dataSourceProviderDefinition, Server targetServer)
        {
            // Get the XMLA Cube definition file
            string[] cubeDbs = Directory.GetFiles(@baseDir + "\\" + sourceServerName + "\\" + sourceDatabaseName + "\\", "*.xmla");
            // Read the XMLA definition into memory and store it within a Database instance                        
            XmlTextReader reader = new XmlTextReader(cubeDbs[0]);
            Database database = new Database();
            Microsoft.AnalysisServices.Utils.Deserialize(reader, database);
            // Set the database name to targetDatabaseName
            database.Name = targetDatabaseName;
            database.ID = targetDatabaseName;
            // Add the unit-test database to the target Analysis server
            targetServer.Databases.Add(database);
            database.Update();
            // Set the Cube's dimension/fact table datasource to the unit-test version.
            foreach (DataSource ds in database.DataSources)
            {
                ds.ConnectionString = dataSourceProviderDefinition;
                ds.Update();
            }

            // Update the data source views
            foreach (DataSourceView dsv in database.DataSourceViews)
            {
                dsv.Update();
            }

            // Update the dimensions
            foreach (Microsoft.AnalysisServices.Dimension dim in database.Dimensions)
            {
                dim.Update();
            }

            // Update the cube
            foreach (Cube cube in database.Cubes)
            {
                cube.Update(UpdateOptions.ExpandFull);
            }
        }

        /** Process all the dimensions within the Database */
        private void ProcessDimensions(Database db)
        {
            foreach (Microsoft.AnalysisServices.Dimension dim in db.Dimensions)
            {
		// Process the dimension and all objects contained within.
                dim.Process(ProcessType.ProcessFull);
            }
        }

        /** Process all the Cube(s) within the database */
        private void ProcessCube(Database db)
        {
            foreach (Cube cube in db.Cubes)
            {
		// Process the cube and all objects contained within.
                cube.Process(ProcessType.ProcessFull);
            }
        }

        /**
         * Creates a new Cube on a given server using a given database based on the XMLA of an existing cube.
         * sourceServerName - The name of the Analysis server where the current cube resides
         * targetServerName - Th name of the Analysis server where the unit-test version of the cube will be placed
         * sourceDatabaseName - The name of the database where the current cube resides
         * targetDatabaseBame - The name of the database where the unit-test version of the cube will be placed
         * dataSourceProviderDefinition - The connection string URL where the test version of the dimension and fact tables reside
         **/
        public void CreateCubeFromXMLA(string sourceServerName, string targetServerName, string sourceDatabaseName, string targetDatabaseName, string dataSourceProviderDefinition)
        {
            // Connect to the unit-test Analysis server
            Server server = ConnectToServer(targetServerName);
            // Drop the target database if it already exists
            DropDatabase(server, targetDatabaseName);

            // Create the unit-test version of the cube on the given target server and database
            AddCubeToDatabase(cubeConfigBaseDir, sourceServerName, sourceDatabaseName, targetDatabaseName, dataSourceProviderDefinition, server);            
            server.Refresh();
            Database db = GetDatabase(server, targetDatabaseName);            
            // Process the dimensions of the new cube
            ProcessDimensions(db);
            // Process the new cube
            ProcessCube(db);
        }
    }
}
